#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "config.h"
#include "imapfilter.h"
#include "pathnames.h"

#ifdef ENCRYPTED_PASSWORDS
#include <openssl/evp.h>
#endif


extern unsigned int flags;
extern account_t *accounts;
extern char *home;

#ifdef ENCRYPTED_PASSWORDS
extern char *passphr;
#endif


int store_passwords(account_t ** accts);


/*
 * Get password from user interactively.
 */
void
get_password(char *passwd, size_t pwlen)
{
	char *c;

	tty_disable_echo();

	if (fgets(passwd, pwlen, stdin))
		if ((c = strchr(passwd, '\n')))
			*c = '\0';

	tty_restore();

	putchar('\n');
}


#ifdef ENCRYPTED_PASSWORDS
/*
 * Encrypt and Base64 encode passwords.  Append the MD5 checksum of the
 * passwords before encrypting them.
 */
int
encrypt_passwords(FILE * fd, account_t ** accts)
{
	int i, n;
	unsigned char iv[EVP_MAX_IV_LENGTH];
	unsigned char *key;
	unsigned char buf[ENCRYPTION_BUF];
	unsigned char ebuf[ENCRYPTION_BUF];
	unsigned char bbuf[ENCRYPTION_BUF];
	unsigned char mdv[EVP_MAX_MD_SIZE];
	int mdl, ebufl, bbufl;
	EVP_CIPHER_CTX ctx;
	EVP_MD_CTX mdctx;
	EVP_ENCODE_CTX bctx;

	key = (unsigned char *)smalloc(EVP_MAX_KEY_LENGTH);

	srandom(time(NULL));

	/* Initialization vector. */
	n = 1 + random() % 100000000, 10;
	snprintf(iv, EVP_MAX_IV_LENGTH, "%08d", n);
	fprintf(fd, "%s\n", iv);

	EVP_CIPHER_CTX_init(&ctx);

	EVP_BytesToKey(EVP_bf_cbc(), EVP_md5(), NULL, passphr, strlen(passphr), 1,
	    key, NULL);

	EVP_DigestInit(&mdctx, EVP_md5());
	EVP_EncryptInit(&ctx, EVP_bf_cbc(), key, iv);
	EVP_EncodeInit(&bctx);

	for (i = 0; accts[i] != NULL; i++) {
		snprintf(buf, ENCRYPTION_BUF, "%s %s %s\n", accts[i]->server,
		    accts[i]->username, accts[i]->password);
		EVP_DigestUpdate(&mdctx, buf, strlen(buf));
		EVP_EncryptUpdate(&ctx, ebuf, &ebufl, buf, strlen(buf));
		EVP_EncodeUpdate(&bctx, bbuf, &bbufl, ebuf, ebufl);

		fwrite(bbuf, sizeof(char), bbufl, fd);
	}

	EVP_DigestFinal(&mdctx, mdv, &mdl);

	xstrncpy(buf, ".\n", ENCRYPTION_BUF - 1);

	/* MD5 checksum of data. */
	for (i = 0; i < mdl; i++)
		snprintf(2 + buf + i * 2, ENCRYPTION_BUF - 3 - i * 2, "%02x",
		    mdv[i]);

	EVP_EncryptUpdate(&ctx, ebuf, &ebufl, buf, strlen(buf));
	EVP_EncodeUpdate(&bctx, bbuf, &bbufl, ebuf, ebufl);
	fwrite(bbuf, sizeof(char), bbufl, fd);

	EVP_EncryptFinal(&ctx, ebuf, &ebufl);

	EVP_EncodeUpdate(&bctx, bbuf, &bbufl, ebuf, ebufl);
	fwrite(bbuf, sizeof(char), bbufl, fd);

	EVP_EncodeFinal(&bctx, bbuf, &bbufl);
	fwrite(bbuf, sizeof(char), bbufl, fd);

	EVP_CIPHER_CTX_cleanup(&ctx);

	return 0;
}


/*
 * Decode (Base64) passwords, decrypt them and verify the MD5 checksum.
 */
int
decrypt_passwords(unsigned char **buf, FILE * fd)
{
	int i, j;
	unsigned char iv[EVP_MAX_IV_LENGTH];
	unsigned char *key;
	unsigned char *c;
	unsigned char ebuf[LINE_MAX];
	unsigned char bbuf[LINE_MAX];
	unsigned char mdv[EVP_MAX_MD_SIZE];
	unsigned char mdc[EVP_MAX_MD_SIZE * 2 + 1];
	int mdl, bufl, ebufl;
	EVP_CIPHER_CTX *ctx;
	EVP_MD_CTX mdctx;
	EVP_ENCODE_CTX bctx;

	j = 1;

	c = *buf = (unsigned char *)smalloc(DECRYPTION_BUF * sizeof(char));
	key = (unsigned char *)smalloc(EVP_MAX_KEY_LENGTH);
	ctx = (EVP_CIPHER_CTX *) smalloc(sizeof(EVP_CIPHER_CTX));

	fgets(bbuf, LINE_MAX, fd);

	memcpy(iv, bbuf, EVP_MAX_IV_LENGTH);

	EVP_CIPHER_CTX_init(ctx);

	EVP_BytesToKey(EVP_bf_cbc(), EVP_md5(), NULL, passphr, strlen(passphr),
	    1, key, NULL);

	EVP_DecryptInit(ctx, EVP_bf_cbc(), key, iv);
	EVP_DecodeInit(&bctx);

	while (fgets(bbuf, LINE_MAX, fd)) {
		EVP_DecodeUpdate(&bctx, ebuf, &ebufl, bbuf, strlen(bbuf));
		if (!EVP_DecryptUpdate(ctx, c, &bufl, ebuf, ebufl))
			goto fail;

		c += bufl;
		*c = '\0';

		if (c - *buf > DECRYPTION_BUF * j - 64) {
			i = c - *buf;
			*buf = (char *)srealloc(*buf, DECRYPTION_BUF * ++j);
			c = *buf + i;
			*c = '\0';
		}
	}

	EVP_DecodeFinal(&bctx, ebuf, &ebufl);
	if (!EVP_DecryptFinal(ctx, c, &bufl))
		goto fail;

	c += bufl;
	*c = '\0';

	/* Calculate the MD5 checksum and check if it is correct. */
	if (!(c = strstr(*buf, "\n.\n")))
		goto fail;

	EVP_DigestInit(&mdctx, EVP_md5());
	EVP_DigestUpdate(&mdctx, *buf, c - *buf + 1);
	EVP_DigestFinal(&mdctx, mdv, &mdl);

	for (i = 0; i < mdl; i++)
		snprintf(mdc + i * 2, EVP_MAX_MD_SIZE * 2 + 1 - i * 2, "%02x",
		    mdv[i]);

	c += 3;

	if (strncmp(c, mdc, 32))
		goto fail;

	EVP_CIPHER_CTX_cleanup(ctx);

	sfree(key);
	sfree(ctx);

	return 0;

fail:
	error("Wrong master passphrase.\n");
	EVP_CIPHER_CTX_cleanup(ctx);
	sfree(*buf);
	sfree(key);
	sfree(ctx);

	return ERROR_DECRYPT;
}


/*
 * Store encrypted passwords to file.
 */
int
store_passwords(account_t ** accts)
{
	char pwfile[PATH_MAX];
	FILE *fd;

	snprintf(pwfile, PATH_MAX, "%s/%s", home, PATHNAME_PASSWORD_FILE);

	create_file(pwfile, S_IRUSR | S_IWUSR);

	fd = fopen(pwfile, "w");

	if (fd == NULL)
		fatal(ERROR_FILE_OPEN, "opening passwords file %s; %s\n",
		    pwfile, strerror(errno));

	encrypt_passwords(fd, accts);

	fclose(fd);

	return 0;
}


/*
 * Interactive encrypted passwords editor.
 */
void
password_editor(void)
{
	int i, q, n, pn;
	char buf[LINE_MAX];
	char *c;
	char *p[2];
	account_t *a, **accts;

	if (!(flags & FLAG_BLANK_PASSWORD)) {
		error("no candidate passwords for encryption found\n");
		return;
	}
	q = pn = 0;

	for (a = accounts; a != NULL; a = a->next)
		if (a->passwdattr == PASSWORD_NONE ||
		    a->passwdattr == PASSWORD_ENCRYPTED)
			pn++;

	accts = (account_t **) xmalloc((pn + 1) * sizeof(account_t *));

	memset(accts, 0, (pn + 1) * sizeof(account_t *));

	for (i = 0, a = accounts; a != NULL; a = a->next) {
		if (a->passwdattr == PASSWORD_NONE ||
		    a->passwdattr == PASSWORD_ENCRYPTED)
			accts[i++] = a;
	}

	do {
		printf("cmd: ");
		fgets(buf, LINE_MAX, stdin);
		c = buf;
		for (;; c++) {
			if (*c == ' ' || *c == '\t')
				continue;
			else if (*c == '?' || *c == 'h')
				printf("c\tclear a password entry\n"
				    "e\tedit a password entry\n"
				    "h\thelp\n"
				    "l\tlist entries\n"
				    "p\tchange master password\n"
				    "q\tquit without saving\n"
				    "w\tsave changes\n"
				    "x\tsave and exit\n");
			else if (*c == 'q')
				q = 1;
			else if (*c == 'l')
				for (i = 0; i < pn; i++)
					printf("%d %s %s %s\n", i + 1,
					    accts[i]->server,
					    accts[i]->username,
					    accts[i]->password);
			else if (*c == 'e') {
				n = atoi(++c);
				if (n == 0 || n < 1 ||
				    n > pn ||
				    accts[n - 1] == NULL)
					break;
				accts[n - 1]->password[0] = '\0';
				printf("Enter new password: ");
				if (fgets(accts[n - 1]->password, PASSWORD_LEN,
					stdin) &&
				    (c = strchr(accts[n - 1]->password, '\n')))
					*c = '\0';
			} else if (*c == 'c') {
				n = atoi(++c);
				if (n == 0 || n < 1 ||
				    n > pn ||
				    accts[n - 1] == NULL)
					break;
				accts[n - 1]->password[0] = '\0';
			} else if (*c == 'p') {
				p[0] = (char *)smalloc(PASSPHRASE_LEN);
				p[1] = (char *)smalloc(PASSPHRASE_LEN);
				do {
					for (i = 0; i < 2; i++) {
						printf("Enter %snew master "
						    "password: ",
						    i ? "again " : "");
						get_password(p[i],
						    PASSPHRASE_LEN);
					}
				} while (strcmp(p[0], p[1]));
				xstrncpy(passphr, p[0], PASSPHRASE_LEN - 1);
				sfree(p[0]);
				sfree(p[1]);
			} else if (*c == 'w' || *c == 's') {
				store_passwords(accts);
			} else if (*c == 'x') {
				store_passwords(accts);
				q = 1;
			} else
				break;
		}
	} while (!q);
}
#endif				/* ENCRYPTED_PASSWORDS */
